/*
	Program that finds triangles in INPUT pgm image
	Outputs points in triangles.

	Sergio Herrera May/2018
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "matrix.h"
#include "pnm.h"
#include "fill.h"
#include <float.h>
#include <math.h>

#include "morphology.h"
#include "follow.h"
#include "chamfer.h"

#define CINC 5

int do_work( char *namein );
void negate( MATRIX* );
void umbral( MATRIX* , int );
void findOBJ( const MATRIX *IMG , int* X, int* Y , int COLOR);
void findFurthestPx( const MATRIX *IMG ,int X, int Y, int* Fx, int* Fy , int COLOR);
void findFurthestVec( const MATRIX *IMG ,int X1, int Y1, int X2 ,int Y2, int* X3, int* Y3 , int COLOR);
void drawPoint( MATRIX *IMG , int X, int Y , int R , int C);
void cleanBorder( MATRIX *IN, int R);

void getPerimeterPCA(MATRIX*, int,int,int,int,int,int);
void ajusta_linea_pca(const int *PX ,const int *PY, int count, float * Point , float * Dir);
void intersect(const float * P1, const float * U1, const float * P2, const float * U2 , float* R);
void sortPointsCCW(const int X1, const int Y1, int* X2, int* Y2,int* X3,int* Y3);

extern MATRIX * convolution ( MATRIX *IN, MATRIX *FIL, int ka );


int main( int argc, char *argv[] )
{
	char namein[64];

	if ( argc < 2 ) { 
		fprintf( stderr, "  [!]Use: %s input_image.pbm \n", argv[0]);
		return -1;
	}

	strncpy( namein, argv[1], 64 );
	
	if ( do_work( namein ) >= 0 ) {
		printf("# Program Done! \n");
		//fprintf( stderr, "# Done!!! \n" );
		return 0;
	}
	else {
		fprintf( stderr, "ERROR: Something is wrong!!!\n" );
		exit(1);
	}

	return 0;
}

int do_work( char *namein ) 
{
	MATRIX *IMG, *IN, *FIL;

	IN = read_pnm( namein ); 

	if ( IN ==NULL ) {
		fprintf( stderr, "ERROR reading %s file\n", namein );
		return -1;
	}
	if ( IN->type != 5 ) {
		fprintf( stderr, "ERROR Bad format. Image must be in b/w\n");
		return -1;
	}
	
	//2 - Smoothing (2 Pass)
		int **fil;
		FIL = matrix_allocate( 3, 3, sizeof( int ) );
		if ( FIL==NULL ) {
			fprintf( stderr, "ERROR allocating memory\n" );
			return -1;
		}
		//Smooth Window
		fil = (int **)FIL->ptr;
		fil[0][0] = 1; fil[0][1] = 1; fil[0][2] = 1;
		fil[1][0] = 1; fil[1][1] = 1; fil[1][2] = 1;
		fil[2][0] = 1; fil[2][1] = 1; fil[2][2] = 1;

		IMG = convolution( IN, FIL, 9 );	//1st Pass
				//write_pnm( IMG, "convol1.pgm" );

		//memcpy( IN, IMG, sizeof(MATRIX) );
		//IMG = convolution( IN, FIL, 9 );	//2nd Pass
				//write_pnm( IMG, "convol2.pgm" );

	
	//3 - Binarize IMG global threshold 
	umbral( IMG, 150 );

	//4 - Opening / Erotion->Dilatation
	int Umbral = 2;
	//Opening
			//Erosion
			negate( IMG );
			chamfer34( IMG ); 
			umbral( IMG , Umbral );
			negate( IMG );
					//write_pnm( IMG, "EROTION.pgm" );
			//Dilatation
			chamfer34( IMG );
			umbral( IMG , Umbral );
					//write_pnm( IMG, "DILATATION.pgm" );
			cleanBorder( IMG , Umbral );
					//write_pnm( IMG, "Opening.pgm" );

	//5 - Extract triangles (Fill)
		printf("#Looking for objects... \n");
		int X,Y;
		int OBJcount = 0;
		int OBJ[5];
		int COLOR;
		
		//Look for objects to fill them
		while(1)
		{
			findOBJ(IMG, &X, &Y, 0);
			if(X > -1){ //Valid Point
				//printf("#Found object at: %d , %d \n", X,Y);
				COLOR = FILLTONE + OBJcount * CINC;
				
				push(X,Y);
				do_fill(IMG, OBJ , 0 , COLOR );	//Fill 0's with COLOR

				OBJcount+=1;
			}
			else break;
		}
		printf("# \t [%d] Objects found! \n", OBJcount);
		if(!OBJcount) return( -1 );

				//write_pnm( IMG, "Segmentation.pgm" );		


	//6 - Compute vertices of each triangle
		printf("# \n" );
		for (int i = 0; i < OBJcount; ++i)
		{
			COLOR = FILLTONE + i * CINC;

			//a) Find first pixel of OBJColor
			findOBJ(IMG, &X, &Y, COLOR); //Find first vertex
			if(X > -1){ //Valid Point
				int X2,Y2, X3,Y3;

				printf("#Object [%d] \n",i+1);
				//printf("Found object at: %d , %d \n", X,Y);
				//printf("Looking for color : %d \n", COLOR);
				findFurthestPx(IMG, X, Y, &X2, &Y2, COLOR); //Find 2nd vertex 
				//printf("Found furthest at: %d , %d \n", X2,Y2);
				findFurthestVec(IMG, X,Y,   X2,Y2 ,    &X3, &Y3,   COLOR); //Find 3rd vertex 
				//printf("Found 3rd at: %d , %d \n", X3,Y3);
				printf("#\t %d,%d \t %d,%d \t %d,%d \t#OLD\n", X,Y,  X2,Y2,  X3,Y3);
				//Three vertices found!


				//Get Perimeter
				push(X,Y);
				do_fill(IMG, OBJ , COLOR , 1 );	//Fill COLOR's with 1's
				//write_pnm( IMG, "Fill.pgm" );		

				dilation_4neig_inplace_obj( IMG , OBJ, 1 );

						//write_pnm( IMG, "perimetroD.pgm" );		

				//Convert perimeter
				byte *pimg;
				int i,j;
				for( i=OBJ[INICIOY]; i<=OBJ[FINY]; i++ ) {
					pimg = (byte *)IMG->ptr[i] + OBJ[INICIOX];
					for( j=OBJ[INICIOX]; j<=OBJ[FINX]; j++ ) {
						if ( *pimg == 254 ) {
							*pimg = 200;
						}
						pimg++;
					}
				}

						//write_pnm( IMG, "perimetroF.pgm" );		

				//Sort points Circula CCW order. 
				sortPointsCCW( X,Y, &X2, &Y2, &X3, &Y3);

				//Follow_perimeter along points
				getPerimeterPCA(IMG,  X,Y, X2,Y2, X3,Y3);

						//write_pnm( IMG, "perimetroG.pgm" );
				
			}
		}

	//7 - Output points

	//if ( write_pnm( IMG, "fillTest1.pgm" ) != 0 )  
    //    return -1;

	matrix_free( IMG );
	matrix_free( IN );
	matrix_free( FIL );

	return 0;
}

// --------------------
// Traverses the perimeter in IMG going through 3 points
void getPerimeterPCA(MATRIX* IMG , int X1 ,int Y1, int X2,int Y2, int X3,int Y3)
{	
	/*
	int a,b,c;

	a = abs(X1-X2);
	b = abs(X1-X3);
	c = abs(X2-X3);

	int dX = (a > b)? ((a>c)?a:c):((b>c)?b:c) ;

	a = abs(Y1-Y2);
	b = abs(Y1-Y3);
	c = abs(Y2-Y3);

	int dY = (a > b)? ((a>c)?a:c):((b>c)?b:c) ;
 
 	int M = dX + dY;
	printf(" MAX difference %d:%d -> %d \n", dX, dY, M);
	*/

	//Get BoundingBox
	//New Method
	int xMin, xMax, yMin, yMax;

	xMax = (X1 > X2)? ((X1>X3)?X1:X3) : ((X2>X3)?X2:X3) ;
	xMin = (X1 < X2)? ((X1<X3)?X1:X3) : ((X2<X3)?X2:X3) ;
	yMax = (Y1 > Y2)? ((Y1>Y3)?Y1:Y3) : ((Y2>Y3)?Y2:Y3) ;
	yMin = (Y1 < Y2)? ((Y1<Y3)?Y1:Y3) : ((Y2<Y3)?Y2:Y3) ;

	int dX = xMax - xMin;
	int dY = yMax - yMin;
 	int Max = dX + dY; //Max difference


	///printf(" MAX difference %d:%d -> %d \n", dX, dY, M);

 	//Allocate memory enough for ANY perimeter segment
	int *pX = malloc( sizeof(int) * Max );
	int *pY = malloc( sizeof(int) * Max );

	int Counter;
	float P1[2], P2[2], P3[2];
	float U1[2], U2[2], U3[2];


	follow_perimeter( IMG , Y1,X1, Y2,X2, 200 , &Counter, pX,pY);
		//printf("The FIRST perimeter segment has [ %d ] px \n", Counter );
		//write_pnm( IMG, "followPer1.pgm" );
	ajusta_linea_pca( pX , pY, Counter , P1 ,  U1 );
	//printf(" Point= [%f , %f] \n", P1[0], P1[1] );
	//printf(" dir= [%f , %f] \n", U1[0], U1[1] );

	
	//printf("\t 2nd\n");
	follow_perimeter( IMG , Y2,X2, Y3,X3, 200 , &Counter, pX,pY);
		//printf("The SECOND perimeter segment has [ %d ] px \n", Counter );
		//write_pnm( IMG, "followPer2.pgm" );
	ajusta_linea_pca( pX , pY, Counter , P2 ,  U2 );
	///printf(" Point= [%f , %f] \n", P2[0], P2[1] );
	///printf(" dir= [%f , %f] \n", U2[0], U2[1] );
	

	//printf("\t 3rd\n");
	follow_perimeter( IMG , Y3,X3, Y1,X1, 200 , &Counter, pX,pY);
		//printf("The THIRD perimeter segment has [ %d ] px \n", Counter );
		//write_pnm( IMG, "followPer2.pgm" );
	ajusta_linea_pca( pX , pY, Counter , P3 ,  U3 );
	///printf(" Point= [%f , %f] \n", P3[0], P3[1] );
	///printf(" dir= [%f , %f] \n", U3[0], U3[1] );

	float nP1[2];
	float nP2[2];
	float nP3[2];

	//Get now
	intersect( P1,U1,  P3,U3,	nP1);
	intersect( P1,U1,  P2,U2, 	nP2);
	intersect( P2,U2,  P3,U3, 	nP3);

	//Print new points
	printf("#\t %.2f,%.2f \t %.2f,%.2f \t %.2f,%.2f #NEW FLOAT\n", 
													nP1[0],nP1[1],
													nP2[0],nP2[1],
													nP3[0],nP3[1]);
	//Round new points
	printf("\t %d,%d \t %d,%d \t %d,%d \n", 
										(int)floor(nP1[0]+0.5),(int)floor(nP1[1]+0.5),
										(int)floor(nP2[0]+0.5),(int)floor(nP2[1]+0.5),
										(int)floor(nP3[0]+0.5),(int)floor(nP3[1]+0.5)  );

	/*
	printf("\t OUTSIDE: Pixels of segment\n");
	for (int k = 0; k <= Counter ; k++)
	{
		printf(" %d,%d \t", pX[k], pY[k] );
	}
	*/

	free(pX); free(pY);
}



// --------------------
//	Finds the intersection between lines P1,P2 with direction U1,U2 respectively.
void intersect(const float * P1, const float * U1, const float * P2, const float * U2 , float* R)
{

	float delta = U1[0]*(-U2[1]) - (-U2[0]*U1[1]);	//   u1(1)*(-u2(2)) -(-u2(1)*u1(2))

	if (delta == 0) printf("\t ERROR!! The lines are parallel... \n");

	float Px = P2[0] - P1[0];
	float Py = P2[1] - P1[1];

	float deltaT = Px * (-U2[1]) - (-U2[0] * Py);

	float t = deltaT/delta;
	
	///printf("delta= %f \t deltaT=%f \t t=%f \n", delta, deltaT, t );
	
	
	//R contains the intersected point
	R[0] = P1[0] + t * U1[0];
	R[1] = P1[1] + t * U1[1];

}


// --------------------
//
void ajusta_linea_pca(const int *PX ,const int *PY, int count , float * Point , float * Dir)
{	
	//printf("\t# Adjusting line...\n");

	int n;
	double mediax, mediay;
	double x, y, a, b, c;
	double vec[2], mat[4];
	double v, vmin, vmax;
	double epsilon; 

	n = 0;
	mediax = mediay = 0.0;
	
	
	for (int k = 0; k <= count ; k++)
	{
		//printf(" %d,%d \t", PX[k], PY[k] );
		mediax += PX[k];
		mediay += PY[k];
		n++;
	}

	mediax /= n;
	mediay /= n;

	a = b = c = 0;
	
	
	for (int k = 0; k <= count; k++) {	
		x = PX[k];
		y = PY[k];
		x -= mediax;
		y -= mediay;
		a += x*x;
		b += x*y;
		c += y*y;
	}
	// La matriz es ahora
	// [ a b ]
	// [ b c ]

	// Calculamos los eigenvalores
	x = a - c;
	v = x * x + 4.0 * b*b;
	v = sqrt(v);

	vmin = ((c + a) - v) / 2.0;
	vmax = ((c + a) + v) / 2.0;


	mat[0] = a - vmin;
	mat[1] = b;
	mat[2] = b;
	mat[3] = c - vmin;

	v = mat[0] * mat[0] + mat[1] * mat[1];
	epsilon = 8.0 * DBL_EPSILON * vmax;

	if (v <= epsilon) {
		vec[0] = 0.0;
		vec[1] = 1.0;
	}
	else {
		v = mat[2] * mat[2] + mat[3] * mat[3];

		if (v <= epsilon) {
			vec[0] = 1.0;
			vec[1] = 0.0;
		}
		else {
			v = sqrt(v);
			vec[0] = mat[1] / v;
			vec[1] = mat[3] / v;
		}
	}

	//printf( "punto: %lf %lf\n", mediax, mediay );
	//printf( "orientación: %lf %lf\n", vec[0], vec[1] );
	Point[0] = mediax;
	Point[1] = mediay;

	Dir[0] = vec[0];
	Dir[1] = vec[1];
	
}


// --------------------
// 
void sortPointsCCW(const int X1, const int Y1, int* X2, int* Y2,int* X3,int* Y3)
{
	//printf("\t 2nd at: %d , %d \n", *X2, *Y2);
	//printf("\t 3rd at: %d , %d \n", *X3, *Y3);

	float Ux = *X2 - X1;
	float Uy = *Y2 - Y1;

	float Vx = *X3 - X1;
	float Vy = *Y3 - Y1;

	float a = Ux * Vy - Uy * Vx;
	//printf(" u=%f,%f   v=%f,%f \n",Ux,Uy,Vx,Vy );
	//printf(" a=     %f    \n", a );

	if( a > 0 ) //Exchange P2 and P3
	{
		//printf("#!! Changing points... \n");
		int sX = *X3;
		int sY = *Y3;

		*X3 = *X2;
		*Y3 = *Y2;
		
		*X2 = sX;
		*Y2 = sY;
	}

	//printf("\t 2nd at: %d , %d \n", *X2, *Y2);
	//printf("\t 3rd at: %d , %d \n", *X3, *Y3);

}

// --------------------
// 
void drawPoint( MATRIX *IMG , int X, int Y , int R , int C) 
{
	byte *px;
	for (int i = Y-R; i < Y+R; i++)
	{	
		px = (byte *) IMG-> ptr[i];
		for (int j = X-R; j < X+R; j++){
			*px++ = C;
		}
	}
}

// --------------------
// Finds furthest Pixel given a vector discibed by P2-P1
void findFurthestVec( const MATRIX *IMG ,int X1, int Y1, int X2, int Y2, int* X3, int* Y3 , int COLOR)
{
	byte *px;
	long maxDist = 0;
	long dist = 0;
	int Vx,Vy, Ux,Uy, Wx,Wy;

	Vx = X2-X1;
	Vy = Y2-Y1;
	Ux = -Vy;
	Uy = Vx;

	*X3 = *Y3= -1; //No OBJ found!

	for (int i = 0; i < IMG-> rows; i++)
	{	
		px = (byte *) IMG-> ptr[i];
		for (int j = 0; j < IMG-> cols; j++){
			if((*px) == COLOR){
				//printf("[%d,%d]= %d  ",i,j,*px);
				Wx = j-X1;
				Wy = i-Y1;
				dist = abs((Ux * Wx) + (Uy * Wy));
				if(dist > maxDist){
					maxDist = dist;
					*Y3=i;
					*X3=j;
				}
			}
			px++;
		}
	}
}


// --------------------
// Finds furthest Pixel given X,Y in IMG
void findFurthestPx( const MATRIX *IMG ,int X, int Y, int* Fx, int* Fy , int COLOR)
{
	byte *px;
	long maxDist = 0;
	long dist = 0;

	*Fy = *Fx= -1; //No OBJ found!

	for (int i = 0; i < IMG-> rows; i++)
	{	
		px = (byte *) IMG-> ptr[i];
		for (int j = 0; j < IMG-> cols; j++){
			if((*px) == COLOR){
				//printf("[%d,%d]= %d  ",i,j,*px);
				dist = (X-j)*(X-j) + (Y-i)*(Y-i);
				if(dist > maxDist){
					maxDist = dist;
					*Fy=i;
					*Fx=j;
				}
			}
			px++;
		}
	}
}


// --------------------
// Find first OBJ pixel in IMG
void findOBJ( const MATRIX *IMG , int* X, int* Y , int COLOR) 
{
	byte *px;
	for (int i = 0; i < IMG-> rows; i++)
	{	
		px = (byte *) IMG-> ptr[i];
		for (int j = 0; j < IMG-> cols; j++){
			if((*px) == COLOR){
				//printf("[%d,%d]= %d  ",i,j,*px);
				*Y=i;
				*X=j;
				return;
			}
			px++;
		}
	}
	*Y=	*X= -1; //No OBJ found!
}

// --------------------
// Negates binary image
void negate( MATRIX *IMG )
{
	byte *px;
	for (int i = 0; i < IMG-> rows; i++)
	{	
		px = (byte *) IMG-> ptr[i];
		for (int j = 0; j < IMG-> cols; j++){
			*px = ((*px) == 255)? 0:255;
			px++;
		}
	}
}

// --------------------
// Clean clean!
void cleanBorder( MATRIX *IN,  int Radius )
{
	int i, j ;
	byte *ptrin;
	/** limpio del borde de la imagen  **/
	
	//Horizontal
    for (int k = 0; k < Radius; ++k)
    {	
    ptrin = (byte *) IN->ptr[0 + k];
    for( j=0; j<IN->cols; j++ ) *ptrin++ = 255;
    
    ptrin = (byte *) IN->ptr[ IN->rows - 1 -k ];
    for( j=0; j<IN->cols; j++ ) *ptrin++ = 255;
    }
	
	//Vertical
	for (int k = 0; k < Radius; ++k)
    {
    for( i=1; i<IN->rows-1; i++ ){
        ptrin = (byte *) IN->ptr[i] + k ;
        *ptrin = 255;
        ptrin = (byte *) IN->ptr[i] + IN->cols - 1 - k ;
        *ptrin = 255;
    }
	}
	
    
}


// --------------------
// Binarizes image given an umbral
void umbral( MATRIX *IMG, int umbral)
{
	byte *px;
	for (int i = 0; i < IMG-> rows; i++)
	{	
		px = (byte *) IMG-> ptr[i];
		for (int j = 0; j < IMG-> cols; j++){
			*px = ((*px) > umbral)? 255:0;
			px++;
		}
	}
}





